Ext.define('App.storage.Storage', {
    singleton: true,
    mixins: {
        observable: 'Ext.util.Observable'
    },
    requires: [
        'App.utility.File'
    ],
    METHOD_EXECUTE: 1,
    METHOD_FETCH: 2,
    METHOD_FETCHROW: 3,
    METHOD_FETCHONE: 4,
    METHOD_PUSH_EVENTS: 5,
    METHOD_PUSH_USERS: 6,
    METHOD_PUSH_EXTRA: 7,
    METHOD_PUSH_GENERATE_XLS: 8,
    METHOD_PUSH_GENERATE_CSV: 9,
    METHOD_PUSH_OPEN_FILE: 10,

    _wildixIntegration: null,
    _isInitialized: false,
    _isLocked: false,
    installedVersion: '',
    appropriateVersion: '1.0.2',
    dbVersion: '11',

    _lockedDbCounter: 0,
    _lockedDbMaxAttempts: 10,
    _socketLimited: true,

    constructor: function() {
        this.getWildixIntegration().on('connecting', this._onConnectingWildixIntegration, this);
        this.getWildixIntegration().on('open', this._onOpenWildixIntegration, this);
        this.getWildixIntegration().on('error', this._onErrorWildixIntegration, this);
        this.getWildixIntegration().on('message', this._onMessageWildixIntegration, this);
        this.getWildixIntegration().on('response', this._onResponseWildixIntegration, this);
        this.getWildixIntegration().on('close', this._onCloseWildixIntegration, this);

        this.getWildixIntegration().on('requesttimeout', this._onErrorRequestTimeoutWildixIntegration, this);
        this.getWildixIntegration().on('response:cdr_view', this._onErrorResponseWildixIntegration, this);
        this.getWildixIntegration().on('message:cdr_view', this._onMessageWIService, this);

        this.proxyUsers  = Ext.create('App.proxy.Users');
        this.proxyGroups = Ext.create('App.proxy.Groups');
        this.proxyTrunks = Ext.create('App.proxy.Trunks');
        this.proxyPbxes  = Ext.create('App.proxy.Pbxes');
        this.proxyEvents = Ext.create('App.proxy.Events');

        this.addEvents('init');
        this.callParent(arguments);
    },

    /*************************************************
     * start local db specific methods
     *************************************************/

    getWildixIntegration: function(){
        if(this._wildixIntegration === null){
            //Rewrite default authorize settings for WService
            Wildix.Settings.app= "CDRView";
            Wildix.Settings.appVersion= CDRViewVersion;
            //Create instance
            this._wildixIntegration = Wildix.IntegrationService.getInstance();
        }
        return this._wildixIntegration;
    },
    isConnected: function(callback){
        if(this.getWildixIntegration().isConnected()){
                return true;
        }
        return false;
    },
    isOSSupported: function(){
        var platform = navigator.platform||'';
        return platform.indexOf('Win')!==-1||platform.indexOf('Mac')!==-1;
    },
    isInstalled: function(){
        return this.getWildixIntegration().isInstalled();
    },
    isLocked: function(){
        return this._isLocked;
    },
    isVersionReceived: function(){
        return !!this.installedVersion;
    },
    isAppropriateVersion: function(){
        if(this.installedVersion){
            var iVers = this.installedVersion.split('.');
            var aVers = this.appropriateVersion.split('.');
            for (var i=0;i<iVers.length;i++){
                if (parseInt(iVers[i]) < parseInt(aVers[i]))
                    return false;
                if (parseInt(iVers[i]) > parseInt(aVers[i]))
                    return true;
            }
            return true;//the same
        }
        return false;
    },
    alwaysTryingToConnect: function(){
        this.getWildixIntegration().getConnection().maxReconnectAttempts = null;
        //this.getWildixIntegration().refresh();
        //this.getWildixIntegration().open();
    },
    close: function(){
        this.getWildixIntegration().getConnection().close();
    },
    /*************************************************
     * end local db specific methods
     *************************************************/

    /***********************************************************
     * start local db events
     ************************************************************/
    _onConnectingWildixIntegration: function(event, wi){

    },
    _onOpenWildixIntegration: function(event, wi){
        var self = this;
        if (!this._isInitialized){
            setTimeout(function(){
                self.getWildixIntegration().send({'message' : 'M_CDR_VIEW','msgdata' : {'serial':Serial,'dbVersion':self.dbVersion,'command' : 'getversion'}}, function(res){
                    self.isBDInUse(function(isLocked){
                        self._isInitialized = true;
                        self._isLocked = isLocked;

                    });
                    if (res && ! res.error && typeof res == 'string' && res.split('.').length == 3)
                        self.installedVersion = res;
                    console.debug(' ON VERSION ', arguments, res);
                });
            },10);
        }
    },
    _onErrorWildixIntegration: function(event, wi){

    },
    _onMessageWildixIntegration: function(event, message){
            //this.trigger(message.type+':'+message.application, this, message.msgdata);
    },
    _onResponseWildixIntegration: function(event, wi, response){
        console.error("_onResponseWildixIntegration", event, wi, response);
    },
    _onErrorResponseWildixIntegration: function(event, response){
        var errorMessage = response && response.msgdata && response.msgdata.error?response.msgdata.error:'';
        if (errorMessage)
            errorMessage = errorMessage.replace('LuaSQL: ','');
        if (['file is encrypted or is not a database','database disk image is malformed'].indexOf(errorMessage)!==-1 && sessionStorage.reinitializePerformed != '1'){
            sessionStorage.reinitializePerformed = '1';
            this.reinitializeDb();
        } else if (['database is locked'].indexOf(errorMessage) !== -1) {
            this._lockedDbCounter++;
            if (this._lockedDbCounter > this._lockedDbMaxAttempts) {
                this.showErrorDialog(response.msgdata.error)
            }
        } else if (response && response.msgdata && response.msgdata.error){
            this.showErrorDialog(response.msgdata.error)
        }
    },
    showErrorDialog: function(error) {
        this.errorDialog = Ext.create('App.view.bootstrap.ErrorCritical', {
            response: error
        });
        this.errorDialog.show();
    },
    _onErrorRequestTimeoutWildixIntegration:function(event, request){
        console.log('Error Request Timeout WildixIntegration', event, request);
        //check if was connected
        if (! this.installedVersion ){
            return;
        }
        this.errorDialog = Ext.create('App.view.bootstrap.ErrorCritical', {
            response: _t('Request timeout')
        });
        this.errorDialog.show();
    },
    _onCloseWildixIntegration: function(event, wi){

    },
    /***********************************************************
     * end local db events
     ************************************************************/
    getWIServiceUrl: function(){
        var platform = navigator.platform||'';
        if (platform.indexOf('Win')!==-1)
            return "https://files.wildix.com/integrations/win/wiservice/SetupWIService.exe";
        if (platform.indexOf('Mac')!==-1)
            return "https://files.wildix.com/integrations/osx/wiservice/WIService.pkg";
        if (platform.indexOf('Linux')!==-1)
            return "#";
        return "";
    },
    isBrowserSupported: function() {
        // Verify Internet Explorer
        if (navigator.userAgent.match(/MSIE/i) != null) {
            /*var versionOfIE = navigator.userAgent.match(/MSIE ([^ \.]+)/i);

            if (versionOfIE != null && parseFloat(versionOfIE[1]) > 6) {
                return true;
            }*/
            return false;
        }

        // Verify Google Chrome
        if (Ext.isChrome) {
            return true;
        }

        // Verify Safari
        if (Ext.isSafari && navigator.platform.indexOf('Win')===-1) {
            return true;
        }

        // Verify firefox
        if (navigator.userAgent.match(/Firefox/i) != null) {
            return true;
        }

        return false;
    },


    lookup: function(check) {
        var self = this;
        if (!check){
            this.alwaysTryingToConnect();
        }
        if (this._isInitialized){
            this.fireEvent('init');
            return;
        }
        if (check*100>5000){
            this._isInitialized = true;
        }


        setTimeout( function() { self.lookup(check?check+1:1); },100);
    },
    isBDInUse: function(callback){
        this.getWildixIntegration().broadcast({'message' : 'M_CDR_VIEW','msgdata':{command:'ping',serial:Serial,'dbVersion':this.dbVersion}}, function(responses){
            var sameDBClientsCount = 0;
            for (var i=0;i<responses.length;i++){
                if (responses[i].command == 'pong' && responses[i].serial == Serial){
                    sameDBClientsCount++;
                }
            }
            if (callback) callback(sameDBClientsCount>0);
        });
    },
    _onMessageWIService: function(event, response){
        if (response.msgdata && response.msgdata.command == 'ping' && response.msgdata.serial == Serial){
            this.getWildixIntegration().broadcast({'message' : 'M_CDR_VIEW','msgdata':{
                    command:'pong',
                    serial:Serial,
                    'dbVersion':this.dbVersion,
                    userAgent:navigator.userAgent,
                    userName:User.name
            },'id':response.id});
        }

    },
    reinitializeDb: function(){
        this.getWildixIntegration().send({'message' : 'M_CDR_VIEW','msgdata' : {'serial':Serial,'dbVersion':this.dbVersion,'command' : 'clear'}}, function(res){
            window.location.reload();
        });
    },
    bulkInsertPromise: function(tableName, values, stDt){
        var self = this;
        var seq = Promise.resolve();
        if (!values || !values.length || !values[0])
            return seq;
        var fields = Object.keys(values[0]);

        var sqlTableFields = 'INSERT OR REPLACE INTO \'' + tableName + '\' (\'' + fields.join('\', \'') +'\') VALUES ';

        var sql = '';
        var nextRow = '';
        var item = null;

        do {(function(){

            if (!sql){
                sql = sqlTableFields;
            } else {
                sql += ' , ';
            }

            item = values.shift();

            nextRow = '('+fields.map(function(field) {
                return  item[field] == null ? 'null' : self.escape(item[field]);
            }).join(',') + ')';


            sql = sql + nextRow;

        })();} while(values.length>0);

        seq = seq.then(Ext.bind(function (sql) {
            return self.executePromise(sql);
        }, seq, [sql])).then(function (result) {
            return result;
        });

        seq = seq.then(function(result){
            return result;
        });

        seq = seq['catch'](function(e){console.debug('bulkInsert error', e , arguments); return e;});

        return seq;
    },
    __quoteRe: new RegExp("'",'g'),
    escape: function (value) {
        var str = ''+value;
        if (!str || str == '') {
            return "''";
        }
        return "'" + str.replace(this.__quoteRe, "''") + "'";
    },
    executePromise: function(sql, scope){
        var self = this;
        return new Promise(function(resolve, reject) {
            self.executeAsync(sql, function(response){
                if (response && response.error){
                    if (self.needShowDbError(response.error)) {
                        reject(Error(response.error));
                    } else {
                        resolve(response);
                    }
                } else {
                    self.resetLockedDbErrorCounter();
                    resolve(response);
                }
            }, scope);
        });
    },
    fetchPromise: function(sql, scope){
        var self = this;
        return new Promise(function(resolve, reject) {
            self.fetchAsync(sql, function(response){
                if (response && response.error){
                    if (self.needShowDbError(response.error)) {
                        reject(Error(response.error));
                    } else {
                        resolve(response);
                    }
                } else {
                    self.resetLockedDbErrorCounter();
                    resolve(response);
                }
            }, scope);
        });
    },
    resetLockedDbErrorCounter: function() {
        this._lockedDbCounter = 0;
    },
    needShowDbError: function(error) {
        if (error === 'LuaSQL: database is locked') {
            return (this._lockedDbCounter > this._lockedDbMaxAttempts);
        }
        return true;
    },
    executeAsync: function(sql, callback, scope, id) {
        scope = scope || this;
        callback = callback || function(){};
        this.getWildixIntegration().send({'message' : 'M_CDR_VIEW','msgdata' : {'serial':Serial,'dbVersion':this.dbVersion,'command' : 'exec', query:sql}}, function(response){
            callback.call(scope, response);
        });
    },
    fetchAsync: function(sql, callback, scope) {
        scope = scope || this;
        callback = callback || function(){};
        this.getWildixIntegration().send({'message' : 'M_CDR_VIEW','msgdata' : {'serial':Serial,'dbVersion':this.dbVersion,'command' : 'query', query:sql}}, function(response){
            if ((response instanceof Object) && !Ext.Object.getKeys(response).length){
                response = [];
            }
            callback.call(scope, response);
        });
    },
    fetchRowAsync: function(sql, callback, scope) {
        var callbackCover = function(response){
            if (response instanceof Array){
                //take first row (if no rows undeined will returned)
                response = response[0];
            }
            callback.call(scope, response);
        }
        this.fetchAsync(sql, callbackCover, scope);

    },
    fetchOneAsync: function(sql, callback, scope, id, options) {
        var callbackCover = function(response){
            var firstColValue = undefined;
            if (response instanceof Array){
                //take first row (if no rows undeined will returned)
                response = response[0];
                if (response != undefined){
                    for (var prop in response){
                        if (response.hasOwnProperty(prop)){
                            firstColValue = response[prop];
                        }
                    }
                }
            }
            callback.call(scope, firstColValue);
        }
        this.fetchAsync(sql, callbackCover, scope, id, options);
    },
    batch: function() {
        var dispatcher, executions = new Array(), requests = new Array(), results = new Array(), callback, scope;
        if (!this.rid) this.rid = 1;
        // collect requests
        for( i=0; i < arguments.length; i++ ) {
            if( Ext.isArray(arguments[i]) ) {
                requests[i] = arguments[i];
                requests[i][3] = this.rid;
                executions[i] = false;

                this.rid++;
                continue;
            }

            break;
        }

        // collect callback & scope
        for( i = arguments.length - 1; i >= 0; i-- ) {
            if( !Ext.isFunction(arguments[i]) && !Ext.isPrimitive(arguments[i]) && i == arguments.length - 1 ) {
                scope = arguments[i];
                continue;
            }
            if( Ext.isFunction(arguments[i]) && (i == arguments.length - 1 || scope ) ) {
                callback = arguments[i];
                continue;
            }
            if( i+1 != requests.length ) {
                if( !callback ) {
                    throw new Error("Invalid callback argument");
                }

                throw new Error("Invalid "+ (i + 1) +" argument");
            }

            break;
        }

        if( !callback ) {
            throw new Error("Invalid callback argument");
        }
        if( !scope ) {
            scope = this;
        }

        dispatcher = function(rid, result) {
            var i, finish = true;

            for( i=0; i < requests.length; i++ ) {
                if( requests[i][3] == rid ) {
                    results[i] = result;
                    executions[i] = true;
                }
            }

            for( i=0; i < executions.length; i++ ) {
                if( !executions[i] ) {
                    finish = false;
                }
            }


            if( finish ) {
                callback.apply(scope, results);
            }
        }

        // execute
        for( var i=0; i < requests.length; i++ ) {
            switch( requests[i][0] ) {
                case this.METHOD_EXECUTE:
                    this.executeAsync(requests[i][1],  Ext.pass(dispatcher, [requests[i][3]], this));
                break;
                case this.METHOD_FETCH:
                    this.fetchAsync(requests[i][1],    Ext.pass(dispatcher, [requests[i][3]], this));
                break;
                case this.METHOD_FETCHROW:
                    this.fetchRowAsync(requests[i][1], Ext.pass(dispatcher, [requests[i][3]], this));
                break;
                case this.METHOD_FETCHONE:
                    this.fetchOneAsync(requests[i][1], Ext.pass(dispatcher, [requests[i][3]], this));
                break;
                default:
                     this.executeAsync(requests[i][1], Ext.pass(dispatcher, [requests[i][3]], this));
                break;
            }
        }
    },
    pushUsers: function(data, isFirstTimeSync, callback, scope) {
        scope = scope || this;
        callback = callback || function(){};
        this.proxyUsers.pushUsers(data, isFirstTimeSync, Ext.bind(callback,scope));
    },
    pushExtra: function(data, callback, scope) {
        var self = this;
        scope = scope || this;
        callback = callback || function(){};

        console.debug('pushExtra', data, self);

        Promise.resolve().then(function(){
            return new Promise(function(resolve){
                self.proxyGroups.pushGroups(data.groups,resolve);
            });
        }).then(function(){
            return new Promise(function(resolve){
                self.proxyTrunks.pushTrunks(data.trunks,resolve);
            });
        }).then(function(){
            return new Promise(function(resolve){
                self.proxyPbxes.pushPbxes(data.pbxes,resolve);
            });
        }).then(function(){
            callback.apply(scope);
        })['catch'](function(e){
            console.error('error during push extra',e);
        });
    },
    pushEvents: function(data, callback, scope) {
        scope = scope || this;
        callback = callback || function(){};
        this.proxyEvents.pushEvents(data,Ext.bind(callback,scope));
    },
    pushXMLGeneration: function (data, callback, scope) {
        var fileName = this.generateReportName(data, '.xlsx');
        App.utility.File.downloadAsXLSX(data, fileName);
        callback.call(scope);
    },
    pushCSVGeneration: function (data, callback, scope) {
        var fileName = this.generateReportName(data, '.csv');
        App.utility.File.downloadAsCSV(data, fileName);
        callback.call(scope);
    },
    generateReportName: function (reportData, format) {
        var reportPrefix = "cdr-view-";
        var reportName = reportData.name || '';
        if (reportName) {
            reportPrefix = reportName.replace(/\W/g, '_') + '_';
        }

        var reportFilterName = reportData.periodFilterName || '';
        if (reportFilterName) {
            reportPrefix = reportPrefix + reportFilterName.toLowerCase().replace(/\W/g, '_') + '_';
        }

        return reportPrefix + Ext.Date.format(new Date(), 'Y-m-d') + format;
    },
    pushOpenFile: function (data, callback, scope, id) {
    },
    dropIndexPromise: function(){
        var self = this;
        return new Promise(function(resolve, reject) {
            self.getWildixIntegration().send({'message' : 'M_CDR_VIEW','msgdata' : {'serial':Serial,'dbVersion':self.dbVersion,'command' : 'dropindex'}}, function(){
               setTimeout(resolve,2000);
            },120000);
        });
    },
    createIndexPromise: function(){
        var self = this;
        return new Promise(function(resolve, reject) {
            self.getWildixIntegration().send({'message' : 'M_CDR_VIEW','msgdata' : {'serial':Serial,'dbVersion':self.dbVersion,'command' : 'createindex'}}, function(){
                resolve();
            },300000);
        });
    },
    chunkData: function(data, byLength) {
      var chunks = [],i = 0,n = data.length;
      if (!byLength)
          byLength = n;
      while (i < n)
        chunks.push(data.slice(i, i += byLength));
      return chunks;
    },
    getBuilder: function() {
        if (isRemoteMode()) {
            return App.storage.QueryConstructorClickHouse;
        }
        return App.storage.QueryConstructor;
    }
} );
